@@ -6,6 +6,7 @@
 #include <linux/list.h>
 #include <linux/binfmts.h>
 #include <linux/mount.h>
+#include "stap_mmap_lock.h"
 #ifndef STAPCONF_TASK_UID
 #include <linux/cred.h>
 #endif
@@ -35,10 +36,10 @@ static atomic_t __stp_attach_count = ATOMIC_INIT (0);
 
 #define debug_task_finder_attach() (atomic_inc(&__stp_attach_count))
 #define debug_task_finder_detach() (atomic_dec(&__stp_attach_count))
-#define debug_task_finder_report()					\
-    (printk(KERN_ERR "%s:%d - attach count: %d, inuse count: %d\n",	\
-	    __FUNCTION__, __LINE__, atomic_read(&__stp_attach_count),	\
-	    atomic_read(&__stp_inuse_count)))
+#define debug_task_finder_report()			  \
+    dbug_task(1, "attach count: %d, inuse count: %d\n",	  \
+	    atomic_read(&__stp_attach_count),		  \
+	    atomic_read(&__stp_inuse_count))
 #else
 #define debug_task_finder_attach()	/* empty */
 #define debug_task_finder_detach()	/* empty */
@@ -86,6 +87,9 @@ struct stap_task_finder_target {
 
 /* public: */
 	pid_t pid;
+	int build_id_len;
+	uint64_t build_id_vaddr;
+	unsigned char *build_id;
 	const char *procname;
         const char *purpose;
 	stap_task_finder_callback callback;
@@ -101,8 +105,90 @@ struct __stp_tf_task_work {
 	struct task_struct *task;
 	void *data;
 	struct task_work work;
+	task_work_func_t func;
 };
 
+static void
+__stp_tf_task_work_free(struct task_work *work)
+{
+	struct __stp_tf_task_work *tf_work =
+		container_of(work, typeof(*tf_work), work);
+
+	_stp_kfree(tf_work);
+}
+
+static void
+__stp_tf_task_worker_fn(struct task_work *work)
+{
+	/* Go through all the queued task workers and dequeue them in order */
+	while (1) {
+		struct __stp_tf_task_work *entry, *tf_work = NULL;
+		unsigned long flags;
+
+		/* Search for task workers for this task */
+		stp_spin_lock_irqsave(&__stp_tf_task_work_list_lock, flags);
+		list_for_each_entry(entry, &__stp_tf_task_work_list, list) {
+			if (entry->task == current) {
+				tf_work = entry;
+				list_del(&tf_work->list);
+				break;
+			}
+		}
+		stp_spin_unlock_irqrestore(&__stp_tf_task_work_list_lock, flags);
+
+		/* No more workers to execute for this task */
+		if (!tf_work)
+			break;
+
+		/* Run the task worker outside the spin lock so it can sleep */
+		tf_work->func(&tf_work->work);
+	}
+
+	/*
+	 * Free the tf_work associated with this task work. It's guaranteed to
+	 * not be on the task work list anymore because we drained all of the
+	 * workers for this task.
+	 */
+	__stp_tf_task_work_free(work);
+	stp_task_work_func_done();
+}
+
+static void
+__stp_tf_init_task_work(struct task_work *work, task_work_func_t func)
+{
+	struct __stp_tf_task_work *tf_work =
+		container_of(work, typeof(*tf_work), work);
+
+	tf_work->func = func;
+	stp_init_task_work(work, __stp_tf_task_worker_fn);
+}
+
+static int
+__stp_tf_task_work_add(struct task_struct *task, struct task_work *work)
+{
+	struct __stp_tf_task_work *tf_work =
+		container_of(work, typeof(*tf_work), work);
+	unsigned long flags;
+	int rc;
+
+	/* Associate this tf_work with the target task it'll run within */
+	tf_work->task = task;
+
+	/*
+	 * Queue the task worker onto the list. Once the tf_work is added to the
+	 * list, the pointer is only safe while the list lock is held. So we
+	 * need to add the task work while we're still holding the list lock.
+	 */
+	stp_spin_lock_irqsave(&__stp_tf_task_work_list_lock, flags);
+	list_add_tail(&tf_work->list, &__stp_tf_task_work_list);
+	rc = stp_task_work_add(task, work);
+	if (rc)
+		list_del(&tf_work->list);
+	stp_spin_unlock_irqrestore(&__stp_tf_task_work_list_lock, flags);
+
+	return rc;
+}
+
 /*
  * Allocate a 'struct task_work' for use.  Internally keeps track of
  * allocated structs for use when shutting down.
@@ -125,72 +211,54 @@ __stp_tf_alloc_task_work(void *data)
 		return NULL;
 	}
 
-	tf_work->task = current;
 	tf_work->data = data;
 
-	// Insert new item onto list.  This list could be a hashed
-	// list for easier lookup, but as short as the list should be
-	// (and as short lived as these items are) the extra overhead
-	// probably isn't worth the effort.
-	stp_spin_lock_irqsave(&__stp_tf_task_work_list_lock, flags);
-	list_add(&tf_work->list, &__stp_tf_task_work_list);
-	stp_spin_unlock_irqrestore(&__stp_tf_task_work_list_lock, flags);
-
 	return &tf_work->work;
 }
 
-/* 
- * Free a 'struct task_work' allocated by __stp_tf_alloc_task_work().
- */
-static void __stp_tf_free_task_work(struct task_work *work)
-{
-	struct __stp_tf_task_work *tf_work, *node;
-	unsigned long flags;
-
-	tf_work = container_of(work, struct __stp_tf_task_work, work);
-
-	// Remove the item from the list.
-	stp_spin_lock_irqsave(&__stp_tf_task_work_list_lock, flags);
-	list_for_each_entry(node, &__stp_tf_task_work_list, list) {
-		if (tf_work == node) {
-			list_del(&tf_work->list);
-			break;
-		}
-	}
-	stp_spin_unlock_irqrestore(&__stp_tf_task_work_list_lock, flags);
-
-	// Actually free the data.
-	_stp_kfree(tf_work);
-}
-
 /* 
  * Cancel (and free) all outstanding task work requests.
  */
 static void __stp_tf_cancel_all_task_work(void)
 {
-	struct __stp_tf_task_work *node;
+	struct __stp_tf_task_work *node, *tmp;
 	unsigned long flags;
 
 	// Cancel all remaining requests.
 	stp_spin_lock_irqsave(&__stp_tf_task_work_list_lock, flags);
-	list_for_each_entry(node, &__stp_tf_task_work_list, list) {
-	    stp_task_work_cancel(node->task, node->work.func);
-	}
-	stp_spin_unlock_irqrestore(&__stp_tf_task_work_list_lock, flags);
-}
-
-static void __stp_tf_free_all_task_work(void)
-{
-	struct __stp_tf_task_work *node;
-	struct __stp_tf_task_work *tmp;
-	unsigned long flags;
-
-	// Free everything.
-	stp_spin_lock_irqsave(&__stp_tf_task_work_list_lock, flags);
+restart:
 	list_for_each_entry_safe(node, tmp, &__stp_tf_task_work_list, list) {
-	    // Remove the item from the list, then free it.
-	    list_del(&node->list);
-	    _stp_kfree(node);
+		struct __stp_tf_task_work *tf_work;
+		struct task_work *work;
+
+		work = stp_task_work_cancel(node->task, node->work.func);
+		if (!work)
+			continue;
+
+		/*
+		 * There can be multiple queued task workers with the same
+		 * worker func, so there's no guarantee that tf_work == node.
+		 * Therefore, we can only free what stp_task_work_cancel() just
+		 * gave us; freeing 'node' would be unsafe.
+		 */
+		tf_work = container_of(work, typeof(*tf_work), work);
+		list_del(&tf_work->list);
+		_stp_kfree(tf_work);
+
+		/*
+		 * If the tf_work we just freed was the next node in the list,
+		 * then we need to restart the list iteration because
+		 * list_for_each_entry_safe() can't cope with the next node
+		 * being freed. We still need to use list_for_each_entry_safe()
+		 * because we need to get through one successful pass through
+		 * the entire list, since it's not guaranteed that this list
+		 * will be empty when this function exits, as there can still be
+		 * active task workers running, which is fine since the
+		 * stp_task_work API will wait for all task workers to finish
+		 * before allowing the module to unload.
+		 */
+		if (tf_work == tmp)
+			goto restart;
 	}
 	stp_spin_unlock_irqrestore(&__stp_tf_task_work_list_lock, flags);
 }
@@ -262,7 +330,7 @@ stap_register_task_finder_target(struct stap_task_finder_target *new_tgt)
 	new_tgt->ops.report_syscall_exit = \
 		&__stp_utrace_task_finder_target_syscall_exit;
 
-	// Search the list for an existing entry for procname/pid.
+	// Search the list for an existing entry for procname/pid/build-id.
 	list_for_each(node, &__stp_task_finder_list) {
 		tgt = list_entry(node, struct stap_task_finder_target, list);
 		if (tgt == new_tgt) {
@@ -277,7 +345,11 @@ stap_register_task_finder_target(struct stap_task_finder_target *new_tgt)
 			/* pid-based target (a specific pid or all
 			 * pids) */
 			|| (new_tgt->pathlen == 0 && tgt->pathlen == 0
-			    && tgt->pid == new_tgt->pid))) {
+			    && tgt->pid == new_tgt->pid)
+			/* buildid-based target */
+			|| (new_tgt->build_id_len > 0
+			    && new_tgt->build_id_len == tgt->build_id_len
+			    && strcmp(new_tgt->build_id, tgt->build_id) == 0))) {
 			found_node = 1;
 			break;
 		}
@@ -532,6 +604,8 @@ __stp_utrace_attach(struct task_struct *tsk,
 	}
 	else {
 		rc = utrace_set_events(tsk, engine, event_flags);
+		dbug_task(2, "utrace_set_events(%lu) returned %d", event_flags,
+			  rc);
 		if (rc == -EINPROGRESS) {
 			/*
 			 * It's running our callback, so we have to
@@ -553,6 +627,7 @@ __stp_utrace_attach(struct task_struct *tsk,
 
 			if (action != UTRACE_RESUME) {
 				rc = utrace_control(tsk, engine, action);
+				dbug_task(2, "utrace_control(%d) returned %d", action, rc);
 				/* If utrace_control() returns
 				 * EINPROGRESS when we're trying to
 				 * stop/interrupt, that means the task
@@ -588,6 +663,8 @@ __stp_call_callbacks(struct stap_task_finder_target *tgt,
 	struct list_head *cb_node;
 	int rc;
 
+	dbug_task(2, "entering tgt=%p tsk=%p pid=%d", tgt, tsk, tsk ? tsk->tgid : -1);
+
 	if (tgt == NULL || tsk == NULL)
 		return;
 
@@ -600,6 +677,13 @@ __stp_call_callbacks(struct stap_task_finder_target *tgt,
 			continue;
 
 		rc = cb_tgt->callback(cb_tgt, tsk, register_p, process_p);
+
+		dbug_task(1, "tgt %s callback returned %d (proc=%s pid=%d, "
+			  "pathlen=%d, engine-attached=%d)",
+			  (cb_tgt->purpose?:""), rc, cb_tgt->procname,
+			  cb_tgt->pid, (int) cb_tgt->pathlen,
+			  cb_tgt->engine_attached);
+
 		if (rc != 0) {
 			_stp_warn("task_finder %s%scallback for task %d failed: %d",
                                   (cb_tgt->purpose?:""), (cb_tgt->purpose?" ":""),
@@ -608,6 +692,38 @@ __stp_call_callbacks(struct stap_task_finder_target *tgt,
 	}
 }
 
+bool
+__verify_build_id(struct task_struct *tsk, unsigned long addr,
+			  unsigned const char *build_id, int build_id_len)
+{
+#define MAX_HEXSTR_LEN 64
+	int i;
+	unsigned char tsk_build_id[MAX_HEXSTR_LEN + 1];
+
+	if (build_id_len > MAX_HEXSTR_LEN)
+		return false;
+
+	for (i = 0; i < build_id_len / 2; i++) {
+		int rc;
+		unsigned char b;
+
+		if ((rc = __access_process_vm_noflush(tsk, addr + i, &b, 1, 0)) != 1)
+			return false;
+
+		tsk_build_id[i * 2]     = "0123456789abcdef"[b >> 4];
+		tsk_build_id[i * 2 + 1] = "0123456789abcdef"[b & 0xf];
+	}
+	tsk_build_id[build_id_len] = '\0';
+
+	if (strcmp(build_id, tsk_build_id)) {
+		dbug_task(2, "target build-id not matched: [%s] @ 0x%lx != [%s]\n",
+			  build_id, addr, tsk_build_id);
+		return false;
+	}
+
+	return true;
+}
+
 static void
 __stp_call_mmap_callbacks(struct stap_task_finder_target *tgt,
 			  struct task_struct *tsk, char *path,
@@ -686,7 +802,7 @@ __stp_call_mmap_callbacks_with_addr(struct stap_task_finder_target *tgt,
 
 	// The down_read() function can sleep, so we'll call
 	// down_read_trylock() instead, which can fail.
-	if (! down_read_trylock(&mm->mmap_sem))
+	if (! mmap_read_trylock(mm))
 		return;
 	vma = __stp_find_file_based_vma(mm, addr);
 	if (vma) {
@@ -704,7 +820,7 @@ __stp_call_mmap_callbacks_with_addr(struct stap_task_finder_target *tgt,
 		// Allocate space for a path
 		mmpath_buf = _stp_kmalloc(PATH_MAX);
 		if (mmpath_buf == NULL) {
-			up_read(&mm->mmap_sem);
+                        mmap_read_unlock(mm);
 			_stp_error("Unable to allocate space for path");
 			return;
 		}
@@ -731,7 +847,7 @@ __stp_call_mmap_callbacks_with_addr(struct stap_task_finder_target *tgt,
 	// At this point, we're done with the vma (assuming we found
 	// one).  We can't hold the 'mmap_sem' semaphore while making
 	// callbacks.
-	up_read(&mm->mmap_sem);
+        mmap_read_unlock(mm);
 		
 	if (mmpath)
 		__stp_call_mmap_callbacks(tgt, tsk, mmpath, dentry, addr,
@@ -826,16 +942,22 @@ __stp_utrace_attach_match_filename(struct task_struct *tsk,
 
 		tgt = list_entry(tgt_node, struct stap_task_finder_target,
 				 list);
-		// If we've got a matching procname or we're probing
-		// all threads, we've got a match.  We've got to keep
-		// matching since a single thread could match a
-		// procname and match an "all thread" probe.
+		// If we've got a matching procname or a matching build-id
+		// or we're probing all threads, we've got a match.  We've
+		// got to keep matching since a single thread could match a
+		// procname/build-id and match an "all thread" probe.
 		if (tgt == NULL)
 			continue;
+                /* buildid-based target ... gets checked in __stp_tf_quiesce_worker */
+                /* procname-based target */
 		else if (tgt->pathlen > 0
 			 && (tgt->pathlen != filelen
 			     || strcmp(tgt->procname, filename) != 0))
+		{
+			dbug_task(2, "target path NOT matched: [%s] != [%s]",
+			          tgt->procname, filename);
 			continue;
+		}
 		/* Ignore pid-based target, they were handled at startup. */
 		else if (tgt->pid != 0)
 			continue;
@@ -860,7 +982,7 @@ __stp_utrace_attach_match_filename(struct task_struct *tsk,
 		// when the thread gets quiesced.
 		rc = __stp_utrace_attach(tsk, &tgt->ops, tgt,
 					 __STP_ATTACHED_TASK_EVENTS,
-					 UTRACE_STOP);
+					 UTRACE_INTERRUPT);
 		if (rc != 0 && rc != EPERM)
 			break;
 		tgt->engine_attached = 1;
@@ -888,7 +1010,8 @@ __stp_utrace_attach_match_tsk(struct task_struct *path_tsk,
 	printk(KERN_ERR "%s:%d entry\n", __FUNCTION__, __LINE__);
 #endif
 	if (path_tsk == NULL || path_tsk->pid <= 0
-	    || match_tsk == NULL || match_tsk->pid <= 0)
+	    || match_tsk == NULL || match_tsk->pid <= 0
+	    || (_stp_target && match_tsk != path_tsk))
 		return;
 
 	// Grab the path associated with the path_tsk.
@@ -943,13 +1066,9 @@ __stp_tf_clone_worker(struct task_work *work)
 	int rc;
 
 	might_sleep();
-	__stp_tf_free_task_work(work);
 	if (atomic_read(&__stp_task_finder_state) != __STP_TF_RUNNING
-	    || current->flags & PF_EXITING) {
-		/* Remember that this task_work_func is finished. */
-		stp_task_work_func_done();
+	    || current->flags & PF_EXITING)
 		return;
-	}
 
 	__stp_tf_handler_start();
 
@@ -964,10 +1083,6 @@ __stp_tf_clone_worker(struct task_work *work)
 	}
 
 	__stp_tf_handler_end();
-
-	/* Remember that this task_work_func is finished. */
-	stp_task_work_func_done();
-	return;
 }
 
 
@@ -996,13 +1111,15 @@ __stp_utrace_task_finder_report_clone(u32 action,
 		_stp_error("Unable to allocate space for task_work");
 		return UTRACE_RESUME;
 	}
-	stp_init_task_work(work, &__stp_tf_clone_worker);
-	rc = stp_task_work_add(child, work);
-	// stp_task_work_add() returns -ESRCH if the task has already
-	// passed exit_task_work(). Just ignore this error.
-	if (rc != 0 && rc != -ESRCH) {
-		printk(KERN_ERR "%s:%d - stp_task_work_add() returned %d\n",
-		       __FUNCTION__, __LINE__, rc);
+	__stp_tf_init_task_work(work, &__stp_tf_clone_worker);
+	rc = __stp_tf_task_work_add(child, work);
+	if (rc) {
+		__stp_tf_task_work_free(work);
+		// stp_task_work_add() returns -ESRCH if the task has already
+		// passed exit_task_work(). Just ignore this error.
+		if (rc != -ESRCH)
+			printk(KERN_ERR "%s:%d - stp_task_work_add() returned %d\n",
+			       __FUNCTION__, __LINE__, rc);
 	}
 
 	__stp_tf_handler_end();
@@ -1163,7 +1280,7 @@ __stp_call_mmap_callbacks_for_task(struct stap_task_finder_target *tgt,
 
 	// The down_read() function can sleep, so we'll call
 	// down_read_trylock() instead, which can fail.
-	if (! down_read_trylock(&mm->mmap_sem)) {
+	if (! mmap_read_trylock(mm)) {
 		_stp_kfree(mmpath_buf);
 		return;
 	}
@@ -1178,7 +1295,7 @@ __stp_call_mmap_callbacks_for_task(struct stap_task_finder_target *tgt,
 
 	// Now allocate an array to cache vma information in.
 	if (file_based_vmas > 0)
-		vma_cache = _stp_kmalloc(sizeof(struct vma_cache_t)
+		vma_cache = _stp_vzalloc(sizeof(struct vma_cache_t)
 					 * file_based_vmas);
 	if (vma_cache != NULL) {
 		// Loop through the vmas again, and cache needed information.
@@ -1217,7 +1334,7 @@ __stp_call_mmap_callbacks_for_task(struct stap_task_finder_target *tgt,
 	// At this point, we're done with the vmas (assuming we found
 	// any).  We can't hold the 'mmap_sem' semaphore while making
 	// callbacks.
-	up_read(&mm->mmap_sem);
+        mmap_read_unlock(mm);
 
 	if (vma_cache) {
 		int i;
@@ -1253,7 +1370,7 @@ __stp_call_mmap_callbacks_for_task(struct stap_task_finder_target *tgt,
 			}
 			vma_cache_p++;
 		}
-		_stp_kfree(vma_cache);
+		_stp_vfree(vma_cache);
 	}
 
 	_stp_kfree(mmpath_buf);
@@ -1268,20 +1385,42 @@ __stp_tf_quiesce_worker(struct task_work *work)
 		(struct stap_task_finder_target *)tf_work->data;
 
 	might_sleep();
-	__stp_tf_free_task_work(work);
 	if (atomic_read(&__stp_task_finder_state) != __STP_TF_RUNNING
-	    || current->flags & PF_EXITING) {
-		/* Remember that this task_work_func is finished. */
-		stp_task_work_func_done();
+	    || current->flags & PF_EXITING)
 		return;
-	}
 
+        /* If we had a build-id based executable probe (so we have a
+         * tgt->build_id) set, we could not check it back in
+         * __stp_utrace_attach_* because we can't do sleepy
+         * access_process_vm() calls from there.  BUt now that we're
+         * in process context, quiesced, finally we can check.  If we
+         * were build-id based, and the build-id does not match, then
+         * we UTRACE_DETACH from this process and skip the callbacks.
+         *
+         * XXX: For processes that do match, we redo this check every
+         * time this callbacks is encountered somehow.  That's
+         * probably unnecessary.
+         */
+        if (tgt->build_id_len > 0) {
+                int ok = __verify_build_id(current,
+                                           tgt->build_id_vaddr,
+                                           tgt->build_id,
+                                           tgt->build_id_len);
+
+                dbug_task(2, "verified buildid-target process pid=%ld ok=%d\n",
+                          (long) current->tgid, ok);
+                if (!ok) {
+                        // stap_utrace_detach (current, & tgt->ops);
+                        return;
+                }
+        } 
+        
 	__stp_tf_handler_start();
 
-	/* Call the callbacks.  Assume that if the thread is a
-	 * thread group leader, it is a process. */
-	__stp_call_callbacks(tgt, current, 1, (current->pid == current->tgid));
- 
+	/* NB make sure we run mmap callbacks before other callbacks
+	 * like 'probe process.begin' handlers so that the vma tracker
+	 * is already initialized in the latter contexts */
+
 	/* If this is just a thread other than the thread group
 	 * leader, don't bother inform map callback clients about its
 	 * memory map, since they will simply duplicate each other. */
@@ -1289,11 +1428,11 @@ __stp_tf_quiesce_worker(struct task_work *work)
 	    __stp_call_mmap_callbacks_for_task(tgt, current);
 	}
 
-	__stp_tf_handler_end();
+	/* Call the callbacks.  Assume that if the thread is a
+	 * thread group leader, it is a process. */
+	__stp_call_callbacks(tgt, current, 1, (current->pid == current->tgid));
 
-	/* Remember that this task_work_func is finished. */
-	stp_task_work_func_done();
-	return;
+	__stp_tf_handler_end();
 }
 
 static u32
@@ -1303,6 +1442,7 @@ __stp_utrace_task_finder_target_quiesce(u32 action,
 {
 	struct task_struct *tsk = current;
 	struct stap_task_finder_target *tgt = engine->data;
+	struct task_work *work;
 	int rc;
 
 	if (atomic_read(&__stp_task_finder_state) != __STP_TF_RUNNING) {
@@ -1343,39 +1483,29 @@ __stp_utrace_task_finder_target_quiesce(u32 action,
 		_stp_error("utrace_set_events returned error %d on pid %d",
 			   rc, (int)tsk->pid);
 
-	if (in_atomic() || irqs_disabled()) {
-		struct task_work *work;
+	/*
+	 * We could be in atomic context in a syscall tracepoint, which runs
+	 * in an RCU read-side critical section. We can't detect this with
+	 * in_atomic() on non-PREEMPT kernels (i.e., CONFIG_PREEMPT=n) so we
+	 * must always use a task worker here because there's no way to tell if
+	 * sleeping is okay.
+	 */
+	work = __stp_tf_alloc_task_work(tgt);
+	if (work == NULL) {
+		_stp_error("Unable to allocate space for task_work");
+		return UTRACE_RESUME;
+	}
+	__stp_tf_init_task_work(work, &__stp_tf_quiesce_worker);
 
-		/* If we can't sleep, arrange for the task to truly
-		 * stop so we can sleep. */
-		work = __stp_tf_alloc_task_work(tgt);
-		if (work == NULL) {
-			_stp_error("Unable to allocate space for task_work");
-			return UTRACE_RESUME;
-		}
-		stp_init_task_work(work, &__stp_tf_quiesce_worker);
-
-		rc = stp_task_work_add(tsk, work);
+	rc = __stp_tf_task_work_add(tsk, work);
+	if (rc) {
+		__stp_tf_task_work_free(work);
 		/* stp_task_work_add() returns -ESRCH if the task has
 		 * already passed exit_task_work(). Just ignore this
 		 * error. */
-		if (rc != 0 && rc != -ESRCH) {
+		if (rc != -ESRCH)
 			printk(KERN_ERR "%s:%d - stp_task_work_add() returned %d\n",
 			       __FUNCTION__, __LINE__, rc);
-		}
-	}
-	else {
-		/* Call the callbacks.  Assume that if the thread is a
-		 * thread group leader, it is a process. */
-		__stp_call_callbacks(tgt, tsk, 1, (tsk->pid == tsk->tgid));
- 
-		/* If this is just a thread other than the thread
-		   group leader, don't bother inform map callback
-		   clients about its memory map, since they will
-		   simply duplicate each other. */
-		if (tgt->mmap_events == 1 && tsk->tgid == tsk->pid) {
-			__stp_call_mmap_callbacks_for_task(tgt, tsk);
-		}
 	}
 
 	__stp_tf_handler_end();
@@ -1466,22 +1596,15 @@ __stp_tf_mmap_worker(struct task_work *work)
 	struct __stp_tf_map_entry *entry;
 
 	might_sleep();
-	__stp_tf_free_task_work(work);
 
 	// See if we can find saved syscall info.
 	entry = __stp_tf_get_map_entry(current);
-	if (entry == NULL) {
-		/* Remember that this task_work_func is finished. */
-		stp_task_work_func_done();
+	if (entry == NULL)
 		return;
-	}
 
 	if (atomic_read(&__stp_task_finder_state) != __STP_TF_RUNNING
 	    || current->flags & PF_EXITING) {
 		__stp_tf_remove_map_entry(entry);
-
-		/* Remember that this task_work_func is finished. */
-		stp_task_work_func_done();
 		return;
 	}
 
@@ -1506,10 +1629,6 @@ __stp_tf_mmap_worker(struct task_work *work)
 	__stp_tf_remove_map_entry(entry);
 
 	__stp_tf_handler_end();
-
-	/* Remember that this task_work_func is finished. */
-	stp_task_work_func_done();
-	return;
 }
 
 static u32
@@ -1521,6 +1640,8 @@ __stp_utrace_task_finder_target_syscall_exit(u32 action,
 	struct stap_task_finder_target *tgt = engine->data;
 	unsigned long rv;
 	struct __stp_tf_map_entry *entry;
+	struct task_work *work;
+	int rc;
 
 	if (atomic_read(&__stp_task_finder_state) != __STP_TF_RUNNING) {
 		debug_task_finder_detach();
@@ -1553,54 +1674,38 @@ __stp_utrace_task_finder_target_syscall_exit(u32 action,
 			    : "UNKNOWN")))),
 		  entry->arg0, rv);
 
-	if (in_atomic() || irqs_disabled()) {
-		struct task_work *work;
-		int rc;
+	/* If this is mmap()/mmap2(), we need to remember the
+	 * return value. We'll use entry->arg0, since
+	 * mmap()/mmap2() doesn't use that info. */
+	if (entry->syscall_no == MMAP_SYSCALL_NO(tsk)
+	    || entry->syscall_no == MMAP2_SYSCALL_NO(tsk)) {
+		entry->arg0 = rv;
+	}
 
-		/* If this is mmap()/mmap2(), we need to remember the
-		 * return value. We'll use entry->arg0, since
-		 * mmap()/mmap2() doesn't use that info. */
-		if (entry->syscall_no == MMAP_SYSCALL_NO(tsk)
-		    || entry->syscall_no == MMAP2_SYSCALL_NO(tsk)) {
-			entry->arg0 = rv;
-		}
-
-		/* If we can't sleep, arrange for the task to truly
-		 * stop so we can sleep. */
-		work = __stp_tf_alloc_task_work(tgt);
-		if (work == NULL) {
-			_stp_error("Unable to allocate space for task_work");
-			__stp_tf_remove_map_entry(entry);
-			__stp_tf_handler_end();
-			return UTRACE_RESUME;
-		}
-		stp_init_task_work(work, &__stp_tf_mmap_worker);
-		rc = stp_task_work_add(tsk, work);
+	/*
+	 * We could be in atomic context in a syscall tracepoint, which runs
+	 * in an RCU read-side critical section. We can't detect this with
+	 * in_atomic() on non-PREEMPT kernels (i.e., CONFIG_PREEMPT=n) so we
+	 * must always use a task worker here because there's no way to tell if
+	 * sleeping is okay.
+	 */
+	work = __stp_tf_alloc_task_work(tgt);
+	if (work == NULL) {
+		_stp_error("Unable to allocate space for task_work");
+		__stp_tf_remove_map_entry(entry);
+		__stp_tf_handler_end();
+		return UTRACE_RESUME;
+	}
+	__stp_tf_init_task_work(work, &__stp_tf_mmap_worker);
+	rc = __stp_tf_task_work_add(tsk, work);
+	if (rc) {
+		__stp_tf_task_work_free(work);
 		/* stp_task_work_add() returns -ESRCH if the task has
 		 * already passed exit_task_work(). Just ignore this
 		 * error. */
-		if (rc != 0 && rc != -ESRCH) {
+		if (rc != -ESRCH)
 			printk(KERN_ERR "%s:%d - stp_task_work_add() returned %d\n",
 			       __FUNCTION__, __LINE__, rc);
-		}
-	}
-	else {
-		if (entry->syscall_no == MUNMAP_SYSCALL_NO(tsk)) {
-			// Call the callbacks
-			__stp_call_munmap_callbacks(tgt, tsk, entry->arg0,
-						    entry->arg1);
-		}
-		else if (entry->syscall_no == MMAP_SYSCALL_NO(tsk)
-			 || entry->syscall_no == MMAP2_SYSCALL_NO(tsk)) {
-			// Call the callbacks
-			__stp_call_mmap_callbacks_with_addr(tgt, tsk, rv);
-		}
-		else {			// mprotect
-			// Call the callbacks
-			__stp_call_mprotect_callbacks(tgt, tsk, entry->arg0,
-						      entry->arg1, entry->arg2);
-		}
-		__stp_tf_remove_map_entry(entry);
 	}
 
 	__stp_tf_handler_end();
@@ -1660,6 +1765,10 @@ stap_start_task_finder(void)
 		rc = __stp_utrace_attach(tsk, &__stp_utrace_task_finder_ops, 0,
 					 __STP_TASK_FINDER_EVENTS,
 					 UTRACE_RESUME);
+
+		dbug_task(2, "__stp_utrace_attach() for pid %d returned %d",
+		          tsk->tgid, rc);
+
 		if (rc == EPERM) {
 			/* Ignore EPERM errors, which mean this wasn't
 			 * a thread we can attach to. */
@@ -1690,8 +1799,11 @@ stap_start_task_finder(void)
 		mmpath = __stp_get_mm_path(tsk->mm, mmpath_buf, PATH_MAX);
 		task_unlock(tsk);
 		if (mmpath == NULL || IS_ERR(mmpath)) {
-			rc = -PTR_ERR(mmpath);
-			if (rc == ENOENT) {
+			rc = PTR_ERR(mmpath);
+			/* If this was our target then it's a fatal error */
+			if (!_stp_target && rc == -ENOENT) {
+				_stp_warn("Unable to get path (error %d) for pid %d",
+					   rc, (int)tsk->pid);
 				rc = 0;	/* ignore ENOENT */
 				continue;
 			}
@@ -1702,7 +1814,7 @@ stap_start_task_finder(void)
 			}
 		}
 
-		/* Check the thread's exe's path/pid against our list. */
+		/* Check the thread's exe's path/pid/build-id against our list. */
 #ifdef STAPCONF_TASK_UID
 		tsk_euid = tsk->euid;
 #else
@@ -1720,11 +1832,16 @@ stap_start_task_finder(void)
 					 struct stap_task_finder_target, list);
 			if (tgt == NULL)
 				continue;
+			/* buildid-based target ... gets checked in __stp_tf_quiesce_worker */
 			/* procname-based target */
-			else if (tgt->pathlen > 0
+			else if (tgt->build_id == 0 && tgt->pathlen > 0
 				 && (tgt->pathlen != mmpathlen
 				     || strcmp(tgt->procname, mmpath) != 0))
+			{
+				dbug_task(2, "target path not matched: [%s] != [%s]",
+					  tgt->procname, mmpath);
 				continue;
+			}
 			/* pid-based target */
 			else if (tgt->pid != 0 && tgt->pid != tsk->pid)
 				continue;
@@ -1746,7 +1863,10 @@ stap_start_task_finder(void)
 			// Set up events we need for attached tasks.
 			rc = __stp_utrace_attach(tsk, &tgt->ops, tgt,
 						 __STP_ATTACHED_TASK_EVENTS,
-						 UTRACE_STOP);
+						 UTRACE_INTERRUPT);
+			dbug_task(2, "__stp_utrace_attach() for %d returned %d", tsk->tgid,
+				  rc);
+
 			if (rc != 0 && rc != EPERM)
 				goto stf_err;
 			rc = 0;		/* ignore EPERM */
@@ -1771,18 +1891,13 @@ stap_task_finder_post_init(void)
 		return;
 	}
 
-#ifdef DEBUG_TASK_FINDER
-	printk(KERN_ERR "%s:%d - entry.\n", __FUNCTION__, __LINE__);
-#endif
+	dbug_task(2, "entry.");
 	rcu_read_lock();
 	do_each_thread(grp, tsk) {
 		struct list_head *tgt_node;
 
 		if (atomic_read(&__stp_task_finder_state) != __STP_TF_RUNNING) {
-#ifdef DEBUG_TASK_FINDER
-			printk(KERN_ERR "%s:%d - exiting early...\n",
-			       __FUNCTION__, __LINE__);
-#endif
+			dbug_task(2, "exiting early...");
 			break;
 		}
 
@@ -1824,6 +1939,10 @@ stap_task_finder_post_init(void)
 					_stp_error("utrace_control returned error %d on pid %d",
 						   rc, (int)tsk->pid);
 				}
+
+				dbug_task(2, "utrace_control(UTRACE_INTERRUPT) for pid %d "
+					  "returned %d (%d)", tsk->pid, rc, -EINPROGRESS);
+
 				utrace_engine_put(engine);
 
 				/* Since we only need to interrupt
@@ -1835,9 +1954,7 @@ stap_task_finder_post_init(void)
 	} while_each_thread(grp, tsk);
 	rcu_read_unlock();
 	atomic_set(&__stp_task_finder_complete, 1);
-#ifdef DEBUG_TASK_FINDER
-	printk(KERN_ERR "%s:%d - exit.\n", __FUNCTION__, __LINE__);
-#endif
+	dbug_task(2, "exit.");
 	return;
 }
 
@@ -1897,12 +2014,6 @@ stap_stop_task_finder(void)
 	 * to wait on any running task_work items. */
 	utrace_exit();
 
-	/* After utrace_exit(), we're *sure* there are no tracepoint
-	 * probes or task work items running or scheduled to be
-	 * run. So, now would be a great time to actually free
-	 * everything. */
-	__stp_tf_free_all_task_work();
-
 #ifdef DEBUG_TASK_FINDER
 	printk(KERN_ERR "%s:%d - exit\n", __FUNCTION__, __LINE__);
 #endif
