@@ -21,6 +21,29 @@ static inline bool atomic_try_cmpxchg(atomic_t *v, int *old, int new)
 }
 #endif
 
+#ifndef STAPCONF_HLIST_ADD_TAIL_RCU
+// Added in linux 4.7, backported to rhel 7, not present in rhel 6
+#define hlist_first_rcu(head)   (*((struct hlist_node __rcu **)(&(head)->first)))
+#define hlist_next_rcu(node)    (*((struct hlist_node __rcu **)(&(node)->next)))
+
+static inline void hlist_add_tail_rcu(struct hlist_node *n,
+					   struct hlist_head *h)
+{
+  struct hlist_node *i, *last = NULL;
+  
+  for (i = hlist_first_rcu(h); i; i = hlist_next_rcu(i))
+    last = i;
+  
+  if (last) {
+    n->next = last->next;
+    n->pprev = &last->next;
+    rcu_assign_pointer(hlist_next_rcu(last), n);
+  } else {
+    hlist_add_head_rcu(n, h);
+  }
+}
+#endif
+
 #ifndef STAPCONF_ATOMIC_FETCH_ADD_UNLESS
 static inline int atomic_fetch_add_unless(atomic_t *v, int a, int u)
 {
@@ -54,6 +77,7 @@ struct __stp_tf_vma_entry {
 	struct task_struct *tsk;
 	unsigned long vm_start;
 	unsigned long vm_end;
+	unsigned long offset;  //offset from base addr of the module
 	char path[TASK_FINDER_VMA_ENTRY_PATHLEN]; /* mmpath name, if known */
 
 	// User data (possibly stp_module)
@@ -87,6 +111,15 @@ __stp_tf_vma_new_entry(void)
 	return entry;
 }
 
+#ifndef kfree_rcu
+static void __stp_tf_vma_free_entry(struct rcu_head *rcu)
+{
+	struct __stp_tf_vma_entry *entry = container_of(rcu, typeof(*entry), rcu);
+
+	kfree(entry);
+}
+#endif
+
 // __stp_tf_vma_put_entry(): Put a specified number of references on the entry.
 static void
 __stp_tf_vma_put_entry(struct __stp_tf_vma_bucket *bucket,
@@ -106,7 +139,11 @@ __stp_tf_vma_put_entry(struct __stp_tf_vma_bucket *bucket,
 	hlist_del_rcu(&entry->hlist);
 	stp_spin_unlock_irqrestore(&bucket->lock, flags);
 
+#ifdef kfree_rcu
 	kfree_rcu(entry, rcu);
+#else
+	call_rcu(&entry->rcu, __stp_tf_vma_free_entry);
+#endif
 }
 
 // stap_initialize_vma_map():  Initialize the free list.  Grabs the
@@ -193,7 +230,8 @@ __stp_tf_get_vma_bucket(struct task_struct *tsk)
 // only from user context.
 static int
 stap_add_vma_map_info(struct task_struct *tsk, unsigned long vm_start,
-		      unsigned long vm_end, const char *path, void *user)
+		      unsigned long vm_end, unsigned long offset,
+		      const char *path, void *user)
 {
 	struct __stp_tf_vma_bucket *bucket = __stp_tf_get_vma_bucket(tsk);
 	struct __stp_tf_vma_entry *entry;
@@ -214,6 +252,7 @@ stap_add_vma_map_info(struct task_struct *tsk, unsigned long vm_start,
 	entry->tsk = tsk;
 	entry->vm_start = vm_start;
 	entry->vm_end = vm_end;
+	entry->offset = offset;
 	entry->user = user;
 
 	path_len = strlen(path);
@@ -227,7 +266,7 @@ stap_add_vma_map_info(struct task_struct *tsk, unsigned long vm_start,
 	}
 
 	stp_spin_lock_irqsave(&bucket->lock, flags);
-	hlist_add_head_rcu(&entry->hlist, &bucket->head);
+	hlist_add_tail_rcu(&entry->hlist, &bucket->head);
 	stp_spin_unlock_irqrestore(&bucket->lock, flags);
 	return 0;
 }
@@ -277,7 +316,7 @@ stap_remove_vma_map_info(struct task_struct *tsk, unsigned long vm_start)
 static int
 stap_find_vma_map_info(struct task_struct *tsk, unsigned long addr,
 		       unsigned long *vm_start, unsigned long *vm_end,
-		       const char **path, void **user)
+		       unsigned long *offset, const char **path, void **user)
 {
 	struct __stp_tf_vma_bucket *bucket;
 	struct __stp_tf_vma_entry *entry;
@@ -287,7 +326,7 @@ stap_find_vma_map_info(struct task_struct *tsk, unsigned long addr,
 
 	bucket = __stp_tf_get_vma_bucket(tsk);
 	entry = __stp_tf_get_vma_map(bucket, tsk, 1, addr >= entry->vm_start &&
-				     addr < entry->vm_end);
+				     addr <= entry->vm_end);
 	if (!entry)
 		return -ESRCH;
 
@@ -295,6 +334,8 @@ stap_find_vma_map_info(struct task_struct *tsk, unsigned long addr,
 		*vm_start = entry->vm_start;
 	if (vm_end)
 		*vm_end = entry->vm_end;
+	if (offset)
+		*offset = entry->offset;
 	if (path)
 		*path = entry->path;
 	if (user)
@@ -379,8 +420,14 @@ stap_find_exe_file(struct mm_struct* mm)
 	// still use our own code. The original get_mm_exe_file() can
 	// sleep (since it calls down_read()), so we'll have to roll
 	// our own.
-#if defined(STAPCONF_DPATH_PATH) && (LINUX_VERSION_CODE >= KERNEL_VERSION(4,1,0))
+#if defined(STAPCONF_GET_MM_EXE_FILE_EXPORTED) && (LINUX_VERSION_CODE >= KERNEL_VERSION(4,1,0))
 	return get_mm_exe_file(mm);
+#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(4,1,0))
+        typedef typeof(&get_mm_exe_file) get_mm_exe_file_fn;
+        if (kallsyms_get_mm_exe_file == NULL)
+          return NULL; /* can't happen; _stp_handle_start would abort before this point */
+        else
+          return (* (get_mm_exe_file_fn) kallsyms_get_mm_exe_file)(mm);
 #else
 	struct file *exe_file = NULL;
 
