@@ -21,9 +21,12 @@
 #include <linux/namei.h>
 #include <linux/delay.h>
 #include <linux/mutex.h>
-#ifdef CONFIG_SECURITY_LOCKDOWN_LSM
+#ifdef STAPCONF_LOCKDOWN_DEBUGFS
 #include <linux/security.h>
 #endif
+#ifdef STAPCONF_514_PANIC
+#include <linux/panic_notifier.h>
+#endif
 #include "../uidgid_compatibility.h"
 
 static int _stp_exit_flag = 0;
@@ -49,7 +52,6 @@ static int _stp_probes_started = 0;
  * transport state flag is atomic. */
 static atomic_t _stp_transport_state = ATOMIC_INIT(_STP_TS_UNINITIALIZED);
 
-static inline int _stp_trylock_inode(struct inode *inode);
 static inline void _stp_lock_inode(struct inode *inode);
 static inline void _stp_unlock_inode(struct inode *inode);
 
@@ -70,8 +72,8 @@ static inline void _stp_unlock_inode(struct inode *inode);
 #include "procfs.c"
 #include "control.c"
 
-static unsigned _stp_nsubbufs = 8;
-static unsigned _stp_subbuf_size = 65536*4;
+static unsigned _stp_nsubbufs = 256;
+static unsigned _stp_subbuf_size = 8 * STP_BUFFER_SIZE; /* 64K */
 
 /* module parameters */
 static int _stp_bufsize;
@@ -115,6 +117,9 @@ static struct timer_list _stp_ctl_work_timer;
 
 static int _stp_transport_fs_init(const char *module_name)
 {
+	// Default to using procfs
+	procfs_p = 1;
+
         // BTW: testing the other !FOO_p first is to protect against repeated
         // invocations of this function with security_locked_down() changing
 #ifdef STAPCONF_LOCKDOWN_DEBUGFS
@@ -123,11 +128,12 @@ static int _stp_transport_fs_init(const char *module_name)
 		dbug_trans(1, "choosing procfs_p=1\n");
         }
 #endif
-        if (!procfs_p) {
-                debugfs_p = 1;
-		dbug_trans(1, "choosing debugfs_p=1\n");
+#ifdef STAPCONF_LOCKDOWN_KERNEL
+        if (!debugfs_p && kernel_is_locked_down ("debugfs")) {
+                procfs_p = 1;
+		dbug_trans(1, "choosing procfs_p=1\n");
         }
-
+#endif
 #ifdef STAP_TRANS_PROCFS
         procfs_p = 1;
         debugfs_p = 0;
@@ -357,6 +363,13 @@ static int _stp_handle_kallsyms_lookups(void)
                 goto err0;
         }
 #endif
+#if !defined(STAPCONF_GET_MM_EXE_FILE_EXPORTED) && (LINUX_VERSION_CODE >= KERNEL_VERSION(4,1,0))
+        kallsyms_get_mm_exe_file = (void*) kallsyms_lookup_name ("get_mm_exe_file");
+        if (kallsyms_get_mm_exe_file == NULL) {
+                _stp_error("Can't resolve get_mm_exe_file!");
+                goto err0;
+        }
+#endif
 #endif
 #if defined(CONFIG_KALLSYMS) && !defined(STAPCONF_KALLSYMS_LOOKUP_NAME_EXPORTED)
         {
@@ -534,8 +547,8 @@ static void _stp_transport_close(void)
 		   current->pid);
 	_stp_cleanup_and_exit(0);
 	_stp_unregister_ctl_channel();
+	_stp_print_cleanup(); /* Requires the transport, so free this first */
 	_stp_transport_fs_close();
-	_stp_print_cleanup();	/* free print buffers */
 	_stp_mem_debug_done();
 
 	dbug_trans(1, "---- CLOSED ----\n");
@@ -547,6 +560,8 @@ static void _stp_transport_close(void)
  */
 static int _stp_transport_init(void)
 {
+	int ret;
+
 	dbug_trans(1, "transport_init\n");
 #ifdef STAPCONF_TASK_UID
 	_stp_uid = current->uid;
@@ -598,20 +613,28 @@ static int _stp_transport_init(void)
 		dbug_trans(1, "Using %d subbufs of size %d\n", _stp_nsubbufs, _stp_subbuf_size);
 	}
 
-	if (_stp_transport_fs_init(THIS_MODULE->name) != 0)
+	ret = _stp_transport_fs_init(THIS_MODULE->name);
+	if (ret)
 		goto err0;
 
 	/* create control channel */
-	if (_stp_register_ctl_channel() < 0)
+	ret = _stp_register_ctl_channel();
+	if (ret < 0)
 		goto err1;
 
 	/* create print buffers */
-	if (_stp_print_init() < 0)
+	ret = _stp_print_init();
+	if (ret < 0) {
+		errk("%s: can't create print buffers!", THIS_MODULE->name);
 		goto err2;
+	}
 
 	/* set _stp_module_self dynamic info */
-	if (_stp_module_update_self() < 0)
+	ret = _stp_module_update_self();
+	if (ret < 0) {
+		errk("%s: can't update dynamic info!", THIS_MODULE->name);
 		goto err3;
+	}
 
 	/* start transport */
 	_stp_transport_data_fs_start();
@@ -634,24 +657,7 @@ err2:
 err1:
 	_stp_transport_fs_close();
 err0:
-	return -1;
-}
-
-/* returns 1 when the lock is successfully acquired, 0 otherwise. */
-static inline int _stp_trylock_inode(struct inode *inode)
-{
-#ifdef STAPCONF_INODE_RWSEM
-	return inode_trylock(inode);
-#else
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)
-	return mutex_trylock(&inode->i_mutex);
-#else
-	/* NB down_trylock() uses a different convention where 0 means
-	 * the lock is successfully acquired.
-	 */
-	return !down_trylock(&inode->i_sem);
-#endif
-#endif
+	return ret;
 }
 
 static inline void _stp_lock_inode(struct inode *inode)
