@@ -13,7 +13,7 @@
  * You should have received a copy of the GNU General Public License
  * along with this program.  If not, see <http://www.gnu.org/licenses/>.
  *
- * Copyright (C) 2016-2019 Red Hat, Inc.
+ * Copyright (C) 2016-2021 Red Hat, Inc.
  *
  */
 
@@ -41,6 +41,7 @@
 #include <sys/mman.h>
 #include <sys/utsname.h>
 #include <sys/resource.h>
+#include <sys/wait.h>
 #include <pwd.h>
 #include "bpfinterp.h"
 #include "../util.h"
@@ -69,18 +70,41 @@ extern "C" {
 
 using namespace std;
 
-static int group_fd = -1;		// ??? Need one per cpu.
-extern "C" { 
-int log_level = 0;
-};
+// XXX declarations from ../staprun/staprun.h required by start_cmd()
+extern "C" {
+int verbose = 0;
+int read_stdin = 0; // TODO not used; stapbpf doesn't (yet) have input.char?
+char *__name__ = NULL; // XXX aliased as module_name below
 int target_pid = 0;
+char *target_cmd = NULL;
+
+/* ../staprun/common.c functions */
+__attribute__ ((format (printf, 1, 2)))
+void eprintf(const char *fmt, ...)
+{
+  va_list va;
+  va_start(va, fmt);
+  // if (use_syslog)
+  // 	vsyslog(LOG_ERR, fmt, va);
+  // else
+  vfprintf(stderr, fmt, va);
+  va_end(va);
+}
+
+/* ../staprun/common.c functions */
+void start_cmd(void);
+int resume_cmd(void);
+};
+
+static int group_fd = -1;		// ??? Need one per cpu.
+static int target_pid_failed_p = 0;
 static int warnings = 1;
 static int exit_phase = 0;
 static int interrupt_message = 0;
 static FILE *output_f = stdout;
 static FILE *kmsg = NULL;
 
-static const char *module_name;
+#define module_name __name__
 static const char *module_basename;
 static const char *script_name; // name of original systemtap script
 static const char *module_license;
@@ -121,6 +145,9 @@ static std::vector<std::string> interned_strings;
 // Table of map id's for statistical aggregates:
 static std::unordered_map<bpf::globals::agg_idx, bpf::globals::stats_map> aggregates;
 
+// Table of foreach loop information:
+static std::vector<bpf::globals::foreach_info> foreach_loop_info;
+
 // XXX: Required static data and methods from bpf::globals, shared with translator.
 #include "../bpf-shared-globals.h"
 
@@ -246,7 +273,7 @@ static std::vector<uprobe_data> uprobes;
 
 // TODO: Move fatal() to bpfinterp.h and replace abort() calls in the interpreter.
 // TODO: Add warn() option.
-static void __attribute__((noreturn))
+static void __attribute__((noreturn))  __attribute__ ((format (printf, 1, 2)))
 fatal(const char *str, ...)
 {
   if (module_name)
@@ -409,7 +436,7 @@ instantiate_maps (Elf64_Shdr *shdr, Elf_Data *data)
           (unsigned long long) rlim_max_orig,
           (unsigned long long) curr_rlimit.rlim_max,
           strerror(errno));
-  if (log_level > 1)
+  if (verbose > 1)
     {
       fprintf(stderr, "increasing map cur resource limit from %llu to %llu\n",
               (unsigned long long) rlim_orig,
@@ -449,7 +476,7 @@ instantiate_maps (Elf64_Shdr *shdr, Elf_Data *data)
           attrs[i].max_entries = ncpus;
         }
 
-      if (log_level > 2)
+      if (verbose > 2)
         fprintf(stderr, "creating map type %u entry %zu: key_size %u, value_size %u, "
                 "max_entries %u, map_flags %u\n", map_type, i,
                 attrs[i].key_size, attrs[i].value_size,
@@ -773,7 +800,7 @@ kprobe_collect_from_syms(Elf_Data *sym_data, Elf_Data *str_data)
       if (syms[i].st_name < str_data->d_size)
 	name = static_cast<char *>(str_data->d_buf) + syms[i].st_name;
       else
-	fatal("symbol %u has invalid string index\n", i);
+	fatal("symbol %zu has invalid string index\n", i);
       maybe_collect_kprobe(name, i, syms[i].st_shndx, syms[i].st_value);
     }
 }
@@ -834,7 +861,7 @@ register_uprobes()
 	  goto fail_n;
 	}
 
-      if (log_level > 1)
+      if (verbose > 1)
         fprintf(stderr, "Associating probe %zu with uprobe %s\n", i, msgbuf);
 
       ssize_t wlen = write(fd, msgbuf, olen);
@@ -953,7 +980,7 @@ register_kprobes()
 	  goto fail_n;
 	}
 
-      if (log_level > 1)
+      if (verbose > 1)
         fprintf(stderr, "Associating probe %zu with kprobe %s\n", i, msgbuf);
       
       ssize_t wlen = write(fd, msgbuf, olen);
@@ -1361,7 +1388,8 @@ init_perf_transport()
       // Create a data structure to track what's happening on each CPU:
       bpf_transport_context *ctx
         = new bpf_transport_context(cpu, pmu_fd, ncpus, map_attrs, &map_fds,
-                                    output_f, &interned_strings, &aggregates, &error);
+                                    output_f, &interned_strings, &aggregates,
+                                    &foreach_loop_info, &error);
       transport_contexts.push_back(ctx);
     }
 
@@ -1384,7 +1412,7 @@ init_perf_transport()
       if (base == MAP_FAILED)
         fatal("error mmapping header for perf_event fd %d\n", pmu_fd);
       perf_headers.push_back((perf_event_mmap_page*)base);
-      if (log_level > 2)
+      if (verbose > 2)
         fprintf(stderr, "Initialized perf_event output on cpu %d\n", cpu);
     }
 }
@@ -1479,6 +1507,7 @@ load_bpf_file(const char *module)
   unsigned script_name_idx = 0;
   unsigned interned_strings_idx = 0;
   unsigned aggregates_idx = 0;
+  unsigned foreach_loop_info_idx = 0;
   unsigned kprobes_idx = 0;
   unsigned begin_idx = 0;
   unsigned end_idx = 0;
@@ -1522,6 +1551,8 @@ load_bpf_file(const char *module)
         interned_strings_idx = i;
       else if (strcmp(shname, "stapbpf_aggregates") == 0)
         aggregates_idx = i;
+      else if (strcmp(shname, "stapbpf_foreach_loop_info") == 0)
+        foreach_loop_info_idx = i;
       else if (strcmp(shname, "version") == 0)
 	version_idx = i;
       else if (strcmp(shname, "maps") == 0)
@@ -1606,6 +1637,25 @@ load_bpf_file(const char *module)
         }
     }
 
+  // PR23478: Initialize table of foreach loop information.
+  if (foreach_loop_info_idx != 0)
+    {
+      uint64_t *foreachtab = static_cast<uint64_t *>(sh_data[foreach_loop_info_idx]->d_buf);
+      unsigned long long foreachtab_size = shdrs[foreach_loop_info_idx]->sh_size;
+      unsigned ofs = 0; unsigned i = 0;
+      while (ofs < foreachtab_size)
+        {
+          bpf::globals::interned_foreach_info ifi;
+          for (unsigned j = 0; j < bpf::globals::n_foreach_info_fields; j++)
+            {
+              ifi.push_back(foreachtab[i+j]);
+            }
+          foreach_loop_info.push_back(bpf::globals::deintern_foreach_info(ifi));
+          i += bpf::globals::n_foreach_info_fields;
+          ofs = sizeof(uint64_t) * i;
+        }
+    }
+
   // Relocate all programs that require it.
   for (unsigned i = 1; i < shnum; ++i)
     {
@@ -1824,7 +1874,7 @@ perf_event_loop(pthread_t main_thread)
 
   for (;;)
     {
-      if (log_level > 3)
+      if (verbose > 3)
         fprintf(stderr, "Polling for perf_event data on %d cpus...\n", n_active_cpus);
       int ready = poll(pmu_fds, n_active_cpus, 1000); // XXX: Consider setting timeout -1 (unlimited).
       if (ready < 0 && errno == EINTR)
@@ -1835,7 +1885,7 @@ perf_event_loop(pthread_t main_thread)
         {
           if (pmu_fds[i].revents <= 0)
             continue;
-          if (log_level > 3)
+          if (verbose > 3)
             fprintf(stderr, "Saw perf_event on fd %d\n", pmu_fds[i].fd);
 
           ready --;
@@ -1970,7 +2020,7 @@ procfs_write_event_loop (procfsprobe_data* data, bpf_transport_context* uctx)
                     path, strerror(errno));
 
         if (bytes_read > 0)
-          msg.append(std::string(buffer_feed));
+          msg.append(buffer_feed, bytes_read);
 
       } while (bytes_read > 0);
 
@@ -2005,7 +2055,7 @@ procfs_cleanup()
   if (procfsprobes.size() > 0 && remove_file_or_dir(dir))
     fprintf(stderr, "WARNING: an error ocurred while deleting a directory (%s). %s.\n", dir, strerror(errno));
 
-  if (log_level)
+  if (verbose)
     fprintf(stderr, "removed fifo directory %s\n", dir);
 }
 
@@ -2036,7 +2086,7 @@ procfs_spawn(bpf_transport_context* uctx)
       if ((mkfifo(path.c_str(), mode) == -1))
         fatal("an error occured while making procfs fifos. %s.\n", strerror(errno));
 
-      if (log_level)
+      if (verbose)
         fprintf(stderr, "created %c fifo %s\n", data->type, path.c_str());
       
       // TODO: Could set the owner/group of the fifo to the effective user.  or real?
@@ -2052,14 +2102,16 @@ procfs_spawn(bpf_transport_context* uctx)
 static void
 usage(const char *argv0)
 {
-  printf("Usage: %s [-v][-w][-V][-h] [-o FILE] <bpf-file>\n"
-	 "  -h, --help       Show this help text\n"
-	 "  -v, --verbose    Increase verbosity\n"
-	 "  -V, --version    Show version\n"
-	 "  -w               Suppress warnings\n"
-	 "  -x pid           Sets the '_stp_target' variable to pid.\n"
-	 "  -o FILE          Send output to FILE\n",
-	 argv0);
+  printf(_("Usage: %s [-v][-w][-V][-h] [-c cmd] [-x pid] [-o FILE] <bpf-file>\n"), argv0);
+  printf(_("-h, --help       Show this help text.\n"
+           "-v, --verbose    Increase verbosity.\n"
+           "-V, --version    Show version.\n"
+           "-w               Suppress warnings.\n"
+           "-c cmd           Command \'cmd\' will be run and stapbpf will\n"
+           "                 exit when it does.  The '_stp_target' variable\n"
+           "                 will contain the pid for the command.\n"
+           "-x pid           Sets the '_stp_target' variable to pid.\n"
+           "-o FILE          Send output to FILE.\n"));
 }
 
 
@@ -2087,6 +2139,41 @@ sigint(int s)
      fatal("error during bpf map update: %s\n", strerror(errno));
 }
 
+// XXX similar to chld_proc in ../staprun/mainloop.c
+void
+sigchld(int signum)
+{
+  int chld_stat = 0;
+  if (verbose > 2)
+    fprintf(stderr, "sigchld %d (%s)\n", signum, strsignal(signum));
+  pid_t pid = waitpid(-1, &chld_stat, WNOHANG);
+  if (pid != target_pid) {
+    return;
+  }
+
+  if (chld_stat) {
+    // our child exited with a non-zero status
+    if (WIFSIGNALED(chld_stat)) {
+      fprintf(stderr, _("WARNING: Child process exited with signal %d (%s)\n"),
+          WTERMSIG(chld_stat), strsignal(WTERMSIG(chld_stat)));
+      target_pid_failed_p = 1;
+    }
+    if (WIFEXITED(chld_stat) && WEXITSTATUS(chld_stat)) {
+      fprintf(stderr, _("WARNING: Child process exited with status %d\n"),
+          WEXITSTATUS(chld_stat));
+      target_pid_failed_p = 1;
+    }
+  }
+
+  // set exit flag
+  int key = bpf::globals::EXIT;
+  int64_t val = 1;
+
+  if (bpf_update_elem
+       (map_fds[bpf::globals::internal_map_idx], &key, &val, 0) != 0)
+     fatal("error during bpf map update: %s\n", strerror(errno));
+}
+
 int
 main(int argc, char **argv)
 {
@@ -2098,64 +2185,76 @@ main(int argc, char **argv)
 
   int rc;
 
-  while ((rc = getopt_long(argc, argv, "hvVwx:o:", long_opts, NULL)) >= 0)
+  while ((rc = getopt_long(argc, argv, "hvVwc:x:o:", long_opts, NULL)) >= 0)
     switch (rc)
       {
       case 'v':
-        log_level++;
+        verbose++;
         break;
       case 'w':
         warnings = 0;
         break;
-
+      case 'c':
+        target_cmd = optarg;
+        break;
       case 'x':
-	target_pid = atoi(optarg);
-	break;
+        target_pid = atoi(optarg);
+        break;
 
       case 'o':
-	output_f = fopen(optarg, "w");
-	if (output_f == NULL)
-	  {
-	    fprintf(stderr, "Error opening %s for output: %s\n",
-		    optarg, strerror(errno));
-	    return 1;
-	  }
-	break;
+        output_f = fopen(optarg, "w");
+        if (output_f == NULL)
+          {
+            fprintf(stderr, "Error opening %s for output: %s\n",
+                    optarg, strerror(errno));
+            return 1;
+          }
+        break;
 
       case 'V':
         printf("Systemtap BPF loader/runner (version %s, %s)\n"
-	       "Copyright (C) 2016-2020 Red Hat, Inc. and others\n" // PRERELEASE
+               "Copyright (C) 2016-2021 Red Hat, Inc. and others\n" // PRERELEASE
                "This is free software; "
-	       "see the source for copying conditions.\n",
-	       VERSION, STAP_EXTENDED_VERSION);
-	return 0;
+               "see the source for copying conditions.\n",
+               VERSION, STAP_EXTENDED_VERSION);
+        return 0;
 
       case 'h':
-	usage(argv[0]);
-	return 0;
+        usage(argv[0]);
+        return 0;
 
       default:
       do_usage:
-	usage(argv[0]);
-	return 1;
+        usage(argv[0]);
+        return 1;
       }
   if (optind != argc - 1)
     goto do_usage;
+  if (target_cmd && target_pid) {
+    fprintf(stderr, _("ERROR: You can't specify the '-c' and '-x' options together.\n"));
+    goto do_usage;
+  }
 
   // Be sure dmesg mentions that we are loading bpf programs:
   kmsg = fopen("/dev/kmsg", "w");
   if (kmsg == NULL)
-    fprintf(stderr, "WARNING: could not open /dev/kmsg for diagnostics: %s\n", strerror(errno));
+    fprintf(stderr, _("WARNING: Could not open /dev/kmsg for diagnostics: %s\n"), strerror(errno));
 
   load_bpf_file(argv[optind]); // <- XXX initializes cpus online, PR24543 initializes default_cpu
   init_internal_globals();
   init_perf_transport();
 
+  // PR25177: For '-c' option, start the command now so its pid is available:
+  if (target_cmd)
+    start_cmd();
+  // XXX Done before begin probes, after load_bpf_file() sets __name__.
+
   // Create a bpf_transport_context for userspace programs:
   unsigned ncpus = map_attrs[bpf::globals::perf_event_map_idx].max_entries;
   bpf_transport_context uctx(default_cpu, -1/*pmu_fd*/, ncpus,
                              map_attrs, &map_fds, output_f,
-                             &interned_strings, &aggregates, &error);
+                             &interned_strings, &aggregates,
+                             &foreach_loop_info, &error);
 
   if (create_group_fds() < 0)
     fatal("Error creating perf event group: %s\n", strerror(errno));
@@ -2177,8 +2276,21 @@ main(int argc, char **argv)
   signal(SIGINT, (sighandler_t)sigint);
   signal(SIGTERM, (sighandler_t)sigint);
 
-  // PR26109: Only continue startup if begin probes did not exit.
+  // perf_event listener is not yet active.
   bool perf_ioc_enabled = false;
+
+  // PR25177: Signal the '-c' command to start running.
+  int target_rc = 0;
+  if (target_cmd)
+    target_rc = resume_cmd();
+  if (target_rc < 0)
+    goto cleanup_and_exit;
+
+  // PR25177: This is to notify when our child process (-c) ends.
+  if (target_cmd)
+    signal(SIGCHLD, (sighandler_t)sigchld);
+
+  // PR26109: Only continue startup if begin probes did not exit.
   if (!get_exit_status()) { // may already be set by begin probe
     // PR22330: Listen for perf_events:
     std::thread(perf_event_loop, pthread_self()).detach();
@@ -2196,6 +2308,7 @@ main(int argc, char **argv)
     pause();
   }
 
+ cleanup_and_exit:
   // Disable the kprobes before deregistering and running exit probes.
   if (perf_ioc_enabled)
     ioctl(group_fd, PERF_EVENT_IOC_DISABLE, 0);
@@ -2216,19 +2329,24 @@ main(int argc, char **argv)
   exit_phase = 1;
   signal(SIGINT, (sighandler_t)sigint); // restore previously ignored signal
   signal(SIGTERM, (sighandler_t)sigint);
+  signal(SIGCHLD, SIG_IGN); // assume child is no longer relevant
 
-  // Run the end probes.
-  if (prog_end && !error)
-    bpf_interpret(prog_end->d_size / sizeof(bpf_insn),
-                  static_cast<bpf_insn *>(prog_end->d_buf),
-                  &uctx);
+  // PR25177: Skip exit probes if target_cmd did not resume correctly.
+  if (target_rc >= 0)
+    {
+      // Run the end probes.
+      if (prog_end && !error)
+        bpf_interpret(prog_end->d_size / sizeof(bpf_insn),
+                      static_cast<bpf_insn *>(prog_end->d_buf),
+                      &uctx);
+
+      // Run the error probes.
+      if (prog_error && error)
+        bpf_interpret(prog_error->d_size / sizeof(bpf_insn),
+                      static_cast<bpf_insn *>(prog_error->d_buf),
+                      &uctx);
+    }
 
-  // Run the error probes.
-  if (prog_error && error) 
-    bpf_interpret(prog_error->d_size / sizeof(bpf_insn),
-                  static_cast<bpf_insn *>(prog_error->d_buf),
-                  &uctx);
-  
   // Clean up transport layer allocations:
   for (std::vector<bpf_transport_context *>::iterator it = transport_contexts.begin();
        it != transport_contexts.end(); it++)
@@ -2240,10 +2358,13 @@ main(int argc, char **argv)
   int error_count = get_error_count();
 
   if (error_count > 0) {
-    // TODO: Need better color configuration.
+    // TODO: Need better color configuration. Borrow staprun's dbug, warn, err macros?
     fprintf(stderr, "\033[0;33m" "WARNING:" "\033[0m" " Number of errors: %d\n", error_count); 
     return 1;
   }
 
+  if (target_pid_failed_p)
+    return 1;
+
   return 0;
 }
