@@ -7,7 +7,7 @@
  * Public License (GPL); either version 2, or (at your option) any
  * later version.
  *
- * Copyright (C) 2007-2017 Red Hat Inc.
+ * Copyright (C) 2007-2021 Red Hat Inc.
  */
 
 #include "staprun.h"
@@ -23,10 +23,6 @@
 #include "../git_version.h"
 #include "../version.h"
 
-#ifndef OPEN_MAX
-#define OPEN_MAX 256
-#endif
-
 /* variables needed by parse_args() */
 int verbose;
 int suppress_warnings;
@@ -234,7 +230,7 @@ void parse_args(int argc, char **argv)
 		case 'V':
                         /* PRERELEASE */
                         printf(_("Systemtap module loader/runner (version %s, %s)\n"
-                              "Copyright (C) 2005-2020 Red Hat, Inc. and others\n"
+                              "Copyright (C) 2005-2021 Red Hat, Inc. and others\n"
                               "This is free software; see the source for copying conditions.\n"),
                             VERSION, STAP_EXTENDED_VERSION);
                         _exit(0);
@@ -515,8 +511,7 @@ static void fatal_handler (int signum)
         int rc;
         char *str = strsignal(signum);
         rc = write (STDERR_FILENO, ERR_MSG, sizeof(ERR_MSG));
-        rc = write (STDERR_FILENO, str, strlen(str));
-        rc = write (STDERR_FILENO, "\n", 1);
+        fprintf(stderr, "%s\n", str);
         (void) rc; /* notused */
 	_exit(1);
 }
@@ -683,6 +678,7 @@ int send_request(int type, void *data, int len)
 
 static int use_syslog = 0;
 
+ __attribute__ ((format (printf, 1, 2)))
 void eprintf(const char *fmt, ...)
 {
 	va_list va;
@@ -701,24 +697,6 @@ void switch_syslog(const char *name)
 	color_errors = 0;
 }
 
-void print_color(const char *type)
-{
-	if (!color_errors)
-		return;
-
-	if (type == NULL) // Reset
-		eprintf("\033[m\033[K");
-	else {
-		char *seq = parse_stap_color(type);
-		if (seq != NULL) {
-			eprintf("\033[");
-			eprintf(seq);
-			eprintf("m\033[K");
-			free(seq);
-		}
-	}
-}
-
 /* Parse SYSTEMTAP_COLORS and returns the SGR parameter(s) for the given
 type. The env var SYSTEMTAP_COLORS must be in the following format:
 'key1=val1:key2=val2:' etc... where valid keys are 'error', 'warning',
@@ -756,41 +734,3 @@ char *parse_stap_color(const char *type)
 
 	return NULL; /* key not found */
 }
-
-void
-closefrom(int lowfd)
-{
-	long fd, maxfd;
-	char *endp;
-	struct dirent *dent;
-	DIR *dirp;
-
-	/* Check for a /proc/self/fd directory. */
-	if ((dirp = opendir("/proc/self/fd"))) {
-		int dir_fd = dirfd(dirp);
-		while ((dent = readdir(dirp)) != NULL) {
-			fd = strtol(dent->d_name, &endp, 10);
-			if (dent->d_name != endp && *endp == '\0'
-			    && fd >= 0 && fd < INT_MAX && fd >= lowfd
-			    && fd != dir_fd)
-				(void) close((int)fd);
-		}
-		(void) closedir(dirp);
-	}
-	else {
-		/*
-		 * Here we fall back on sysconf(). Why? It is possible
-		 * /proc isn't mounted, we're out of file descriptors,
-		 * etc., which could cause the opendir() to fail. Also
-		 * note thet it is possible to open a file descriptor
-		 * and then drop the rlimit such that it is below the
-		 * open fd.
-		 */
-		maxfd = sysconf(_SC_OPEN_MAX);
-		if (maxfd < 0)
-			maxfd = OPEN_MAX;
-
-		for (fd = lowfd; fd < maxfd; fd++)
-			(void) close((int) fd);
-	}
-}
