@@ -50,6 +50,10 @@ public:
   void emit_module_decls (systemtap_session& s);
   void emit_module_init (systemtap_session& s);
   void emit_module_exit (systemtap_session& s);
+
+  friend void warn_for_bpf(systemtap_session& s,
+                           netfilter_derived_probe_group *dpg,
+                           const std::string& kind);
 };
 
 struct netfilter_var_expanding_visitor: public var_expanding_visitor
@@ -313,7 +317,7 @@ netfilter_derived_probe_group::emit_module_decls (systemtap_session& s)
       s.op->newline() << "#elif defined(STAPCONF_NETFILTER_V41)";
       s.op->newline() << "int (*nf_okfn)(struct sock *, struct sk_buff *) = nf_state->okfn;";
       s.op->newline() << "#endif";
-      common_probe_entryfn_prologue (s, "STAP_SESSION_RUNNING", "stp",
+      common_probe_entryfn_prologue (s, "STAP_SESSION_RUNNING", "", "stp",
                                      "stp_probe_type_netfilter",
                                      false);
 
@@ -417,6 +421,21 @@ netfilter_derived_probe_group::emit_module_exit (systemtap_session& s)
 }
 
 
+// PR26234: Not supported by stapbpf.
+void
+warn_for_bpf(systemtap_session& s,
+             netfilter_derived_probe_group *dpg,
+             const std::string& kind)
+{
+  for (unsigned i=0; i<dpg->probes.size(); i++)
+    {
+      s.print_warning(_F("%s will be ignored by bpf backend",
+                         kind.c_str()),
+                      dpg->probes[i]->tok);
+    }
+}
+
+
 netfilter_var_expanding_visitor::netfilter_var_expanding_visitor (systemtap_session& s):
   var_expanding_visitor (s)
 {
