@@ -1,5 +1,5 @@
 void foo(void) {
-    asm("movq $0x56781234beefdead, %rbx");
+    asm("movl $0xbeefdead, %ebx");
     return;
 }
 
